/*
 *  Copyright (c) 1996 Rainbow Technologies Inc, Internet Security Group. 
 *  All rights reserved.
 *
 *  Copyright (C) 2000 Paolo Molaro <lupus@linuxcare.com>. All rights reserved.
 *  Copyright (C) 2000 Marco Pantaleoni <panta@linuxcare.com>. All rights reserved.
 *
 *  name:           cspci.c
 *
 *  Description:    Source file for CryptoSwift Linux driver
 *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL RAINBOW
 *   TECHNOLOGIES BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN 
 *   AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION 
 *   WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *    
 *   This source code is subject to the U.S. Export Administration Regulations and 
 *   other U.S. law, and may not be exported or re-exported to certain countries 
 *   (currently Afghanistan (Taliban-controlled areas), Cuba, Iran, Iraq, Libya, 
 *   North Korea, Serbia (except Kosovo), Sudan and Syria) or to persons or entities 
 *   prohibited from receiving U.S. exports (including Denied Parties, entities on 
 *   the Bureau of Export Administration Entity List, and Specially Designated 
 *   Nationals). 
 * 
 *   Permission is hereby granted, free of charge, to any person meeting the export 
 *   regulations above to deal in this software and associated documentation files 
 *   (the "Software"), without restriction, including without limitation the rights 
 *   to use, copy, modify, and merge copies of the Software, and to permit persons 
 *   to whom the Software is furnished to do so, subject to the following conditions:
 * 
 *    1.  The above copyright notice and this permission notice shall be included in all 
 *    copies or substantial portions of the Software.
 *
 * $Archive::  $
 */

#define CS_BROKEN_BH 0

#define CS_BUS_MASTER

#include <linux/config.h>
#ifdef CONFIG_SMP
# define __SMP__
#endif

#include <linux/version.h>

#ifdef MODULE
# ifdef MODVERSIONS
#  include <linux/modversions.h>
# endif /* MODVERSIONS */
# include <linux/module.h>
#else
# define MOD_INC_USE_COUNT
# define MOD_DEC_USE_COUNT
# define MOD_IN_USE	1
#endif /* MODULE */

#include <linux/major.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/string.h>
#include <linux/delay.h>

#include <linux/poll.h>

#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/wait.h>
#include <linux/timer.h>
#include <asm/bitops.h>
#include <asm/io.h>

#ifdef HAVE_SHARED_IRQ
#define USE_SHARED_IRQ
#include <linux/shared_irq.h>
#endif

#define LINUX
#include "cspci.h"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#  define KERNEL_2_6
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#  define KERNEL_2_4
#else
# define pci_enable_device(x)       0
# define pci_resource_start(dev,bar) \
     (((dev)->base_address[(bar)] & PCI_BASE_ADDRESS_SPACE) ? \
     ((dev)->base_address[(bar)] & PCI_BASE_ADDRESS_IO_MASK) : \
     ((dev)->base_address[(bar)] & PCI_BASE_ADDRESS_MEM_MASK))
# ifndef module_init
#  define module_init(x)  int init_module(void) { return x(); }
# endif /* module_init */
# ifndef module_exit
#  define module_exit(x)  void cleanup_module(void) { x(); }
# endif /* module_exit */
# ifndef init_waitqueue_head
#  define init_waitqueue_head(q) *q = NULL
# endif /* init_waitqueue_head */
#endif


static void printhex(unsigned char *, unsigned char *, unsigned int);

#define CSPCI_DRIVER_VERSION CS_VERSION_CODING(3,1,8)
static char          *version     = "CryptoSwift Linux Driver v3.1.8";
MODULE_DESCRIPTION ("CryptoSwift Linux Driver v3.1.8");
MODULE_LICENSE ("GPL");

MODULE_AUTHOR ("Rainbow Technologies");
static unsigned long  version_num = CSPCI_DRIVER_VERSION;

#define OPREG_SIZE             64
#define CONTROLREG_SIZE        16
#define TIME_OUT               1000   /* in jiffies (each jiffie=10ms)*/

/* Device Flags */
#define BUS_MASTER_FLAG        0x00000001
#define OUT_OF_ORDER           0x00000002
#define INTERRUPT              0x00000004
#define CARDBUSY               0x00000008
#define USE_OPREG              0x00000010
#define NEED_DPC               0x00000020
#define USE_FASTMAP_IOREG      0x00000040
#define WAIT_FOR_RESET	       0x00000100

/* Status flags */
#define HS_PROCESSING          0x00000001
#define HS_COMPLETE            0x00000002
#define HS_CANCELLABLE         0x00000004
#define HS_CANCELPENDING       0x00000008

/* Register flags, masks, vlaues */
#define VAL_CLEAR              0x00000000
#define VAL_RESET              0x0F000000
#define VAL_ALL_MAIL_EMPTY     0x000F0000
#define VAL_INTERRUPTED        0x00800000
#define VAL_INTCLEAR_BETA      0x203F1000
#define VAL_INTCLEAR_V1        0x023F1000
#define VAL_BMRESET            0x06000000
#define VAL_BMENABLE           0x00004400
#define VAL_CARDBUSY           0x00000002
#define VAL_CS3_ICR_INTENABLE  0x00001020

/* Operational register offsets */
#define REG_COMMAND         0x00
#define REG_FASTMAP_FROM    0x04
#define REG_FASTMAP_TO      0x08
#define REG_SIZES           0x0c
#define REG_COMPCMD         0x10
#define REG_IN_MAILBOX      0x14
#define REG_OUT_MAILBOX     0x18
#define REG_SIGNATURE       0x1c
#define REG_TO_ADDR         0x24
#define REG_TO_SIZE         0x28
#define REG_FROM_ADDR       0x2c
#define REG_FROM_SIZE       0x30
#define REG_MAILSTAT        0x34
#define REG_INTCONT         0x38
#define REG_CONTROL         0x3c

/* Processor register offsets */
#define OPREG_HANDSHAKE       0x4
#define OPREG_STATE           0xA

/* Reading and writing PCI register */
#define CS_W_CONTROL(dev, addr, val) do { outl_p((val),((dev)->controlRegs)+(addr)); } while (0)
#define CS_R_CONTROL(dev, addr)      (inl_p(((dev)->controlRegs)+(addr)))

#define CS_W_OP(dev, addr, val)      do { outb_p((val), ((dev)->opRegs+(addr))); } while (0)
#define CS_R_OP(dev, addr)           (inb_p((dev)->opRegs+(addr)))

#ifdef CS_DEBUG
#define dbg_printf(a)  do {printk a;} while(0)
#else
#define dbg_printf(a)
#endif

/* PCI vendor ID and device ID for CryptoSwift */
#include "csdevid.h"

typedef struct _HANDLEINFO
{
	CS_CMDINFO   cmd_info;
	unsigned int response;
	unsigned int flags;
#ifdef KERNEL_2_4
	wait_queue_head_t waitq;
#else
	struct wait_queue *waitq;
#endif
	SW_BYTE *orig_buf;
	SW_BYTE *buffer;
	caddr_t  dma_buf;					/* sal */

	struct _HANDLEINFO *next;
	unsigned int bufsize;
} HANDLEINFO;

struct cspci_device
{
	int             opRegs;
	int             controlRegs;
	int             irq;
	int             response;
	unsigned int    mem_start;
	struct cspci_device  *next_module;
	unsigned int    flags;
	HANDLEINFO     *ph;					/* handle to the host info */
	int             device_id;
	unsigned int    val_intclear;
	HANDLEINFO     *wait_head;
	HANDLEINFO     *wait_tail;
	int             unit;
	spinlock_t	lock;					/* spin lock for modifying device fields */
	char            name[8];
	char            product_name[60];
	CS_ADAPTERSTATS stat;  /* statistics for adapter */
};

#define WINDOW_SIZE    4*1024

static CS_GLOBALSTATS stat;

static HANDLEINFO *wait_head = NULL;
static HANDLEINFO *wait_tail = NULL;

static spinlock_t cs_glock = SPIN_LOCK_UNLOCKED;

/* Function Prototypes */
int         cs_init(void);
static void cs_cleanup(void);
static int  cs_scan(int do_probe);
static int  cs_found_device(struct pci_dev *dev);
static int  cs_probe1(struct cspci_device *dev);
static int  cs_reset(struct cspci_device *dev);
static int  cs_open(struct inode *inode, struct file *file);
static int  cs_ioctl(struct inode *inode, struct file *file, 
					 unsigned int cmd, unsigned long arg);
static int  cs_release(struct inode *inode, struct file *file);
#ifdef KERNEL_2_6
static irqreturn_t cs_interrupt(int irq, void *dev_id, struct pt_regs *regs);
#else
static void cs_interrupt(int irq, void *dev_id, struct pt_regs *regs);
#endif
static struct cspci_device *cs_requestdev(unsigned long accnum);
static void cs_freedev(struct cspci_device *dev);
static void cs_processreq(unsigned long accnum);
static void cs_executereq(struct cspci_device *dev, HANDLEINFO *ph);
static HANDLEINFO *cs_removehead(struct cspci_device *dev);
static void  cs_queuereq(HANDLEINFO *ph);
static int  cs_dequeuereq(HANDLEINFO *ph);
void        cleanup_module(void);
static void cs_cancel(HANDLEINFO *ph);

#ifdef CS_DEBUG
static void cs_debug(void);
#endif

/* bottom half or DPC */
#if CS_BROKEN_BH
static void cs_bottomentry(void);
#else
static void cs_bottomentry(void *userdata);
#endif
static void cs_dpc(struct cspci_device* dev);
static int  bh_csno = -1;


#if ! CS_BROKEN_BH
#ifdef KERNEL_2_6 
static struct work_struct cs_task;				/* initialized by cs_init                        */
#else
static struct tq_struct cs_task;				/* initialized by cs_init                        */
#endif
#endif
static int bh_scheduled = 0;	/* cs_bottomentry scheduled to run */

/* Device Driver function pointers */
static struct file_operations cs_fops =
{
/*
#ifdef KERNEL_2_4
	owner:   THIS_MODULE,
#endif
*/
	ioctl:   cs_ioctl,
	open:    cs_open,
	release: cs_release
};


/* A list of all installed CryptoSwift devices. Used for removing the driver */
static struct cspci_device **root_cs_dev = NULL;
static int             next_cs_dev = 0;	/* next device to be scaned             */
static int             num_cs_dev = 0;	/* number of use in root_cs_dev[] array */


static int __init cspci_init_module(void)
{
		printk("calling cs_init!\n");
	return cs_init();
}

static void __exit cspci_cleanup_module(void)
{
#ifndef KERNEL_2_6
	if (MOD_IN_USE)
		printk("CSPCI module in use. Cleanup Delayed!\n");
	else
#endif
		cs_cleanup();
}

module_init(cspci_init_module);
module_exit(cspci_cleanup_module);

int cs_init(void)
{
	int cards_found;

#ifndef KERNEL_2_6
	MOD_INC_USE_COUNT;
#endif
	memset( &stat, 0, sizeof(stat) );

	/* copyright message */
	printk("Rainbow Technologies %s\n", version);

	cards_found = cs_scan(0);
	if (cards_found <= 0) {
		dbg_printf(("cs_init: No cards found\n"));
#ifndef KERNEL_2_6
		MOD_DEC_USE_COUNT;
#endif
		return -EIO;
	}

	/* alloc the array for holding all the devices */
	root_cs_dev = (struct cspci_device**)kmalloc( sizeof(struct cspci_device*) * cards_found,
											GFP_KERNEL );
	if (NULL == root_cs_dev) {
#ifndef KERNEL_2_6
		MOD_DEC_USE_COUNT;
#endif
		return -ENOMEM;
	}
	memset( root_cs_dev, 0, sizeof(struct cspci_device*) * cards_found );

#if CS_BROKEN_BH
	for (bh_csno = 31; bh_csno >= 0; --bh_csno)
	{
		if (bh_base[bh_csno] == NULL)
			break;
	}

	if (bh_csno >= 0)
	{

#ifdef KERNEL_2_6
	        INIT_WORK(&cs_task,(void (*)(void *)) cs_bottomentry , NULL);
#else
		init_bh(bh_csno, cs_bottomentry);
#endif
	}
	else
	{
		printk( "cspci: driver is installed without bottom half.\n" );
	}
#else /* if CS_BROKEN_CS*/
#ifndef KERNEL_2_6
	bh_csno = IMMEDIATE_BH;
#ifdef KERNEL_2_4
	cs_task.list.next    = NULL; 
#endif

	cs_task.sync    = 0;
	cs_task.routine = (void (*)(void *)) cs_bottomentry;	/* set bottom half to call later */
	cs_task.data    = NULL;									/* currently UNUSED */
#else /*ifndef KERNEL_2_6*/
        INIT_WORK(&cs_task,(void (*)(void *)) cs_bottomentry , NULL);
#endif /*ifndef KERNEL_2_6*/
#endif /* if CS_BROKEN_CS*/
	cards_found = cs_scan(1);
	if (cards_found <= 0)
	{
		dbg_printf(("cs_init: No cards initialized\n"));
		kfree(root_cs_dev);
#ifndef KERNEL_2_6
		MOD_DEC_USE_COUNT;
#endif
		return -EIO;
	}

	stat.driverVersion  = version_num;
	stat.numAccelerator = cards_found;

	/* Register the device */
	if (register_chrdev(CS_MAJOR,"cspci",&cs_fops))
	{
		dbg_printf(("Could not get major number %d for CSPCI device driver\n", CS_MAJOR)); 
		cs_cleanup();
#ifndef KERNEL_2_6
		MOD_DEC_USE_COUNT;
#endif
		return -EIO;
	}

#ifndef KERNEL_2_6
  	MOD_DEC_USE_COUNT;
#endif
	return cards_found < 0 ? cards_found : 0;

}

static void cs_cleanup(void)
{
	int i;

	unregister_chrdev(CS_MAJOR, "cspci");

  	for (i = 0; i < num_cs_dev; ++i)
	{
		struct cspci_device* dev;
		dev = root_cs_dev[i];
		if (dev != NULL)
		{
			free_irq(dev->irq, (void*)dev);
			if (dev->flags & USE_OPREG)
				release_region(dev->opRegs, OPREG_SIZE);
			release_region(dev->controlRegs, CONTROLREG_SIZE);
			kfree(dev);
			dbg_printf(("Freeing CS resources\n"));
		}
		root_cs_dev[i] = NULL;
	}

	kfree(root_cs_dev);
	root_cs_dev = NULL;

#if CS_BROKEN_BH
	if (bh_csno >= 0)
	{
		bh_base[bh_csno] = NULL;
		bh_csno = -1;
	}
#endif

	dbg_printf(("Done CSPCI\n\n"));
}

static int cs_scan(int do_probe)
{
	int cards_found = 0;
	unsigned char  pci_irq_line; /*, pci_latency; */
	unsigned int   pci_ioaddr0, pci_ioaddr1;
	unsigned short pci_command;
/*  int pci_index; */
	unsigned char fun;					/* deviceID function */
	unsigned short ver;					/* deviceID versions */
	int rc = 0;
	struct pci_dev  *cspci_dev = NULL;

	dbg_printf(("cs_scan: Top of function\n"));
#ifndef KERNEL_2_6
	if (!pci_present())
		return cards_found;
#endif
	for (fun = 0; fun <= DEVID_FUN_MAX; ++fun)
	{
		for (ver = 0; ver <= DEVID_VER_MAX; ++ver)
		{
			unsigned int device_id;
			device_id = DEVID_CONCAT(DEVID_ICD_PUBLIC, fun, ver);

/*          pci_index = 0; */
			/* query PCI BIOS for appropriate VendorID and DeviceID */
			while ((cspci_dev = pci_find_device( ISG_VENDOR_ID,
												 device_id,
												 cspci_dev )))
			{
				dbg_printf(("cs_scan: found device 0x%04x\n", device_id));
dbg_printf(("cs_scan: found device 0x%4x,0x%04x\n", ISG_VENDOR_ID, device_id));

				/* increment the index so that next pcibios_find_device() will
				   return the next device */
				/*       pci_index++; */ /* Andrew */

   				/* maybe do this only on probe? - lupus */
				dbg_printf(("cs_scan: enabling device\n"));
				if (pci_enable_device(cspci_dev) < 0) {
					dbg_printf(("cs_scan: device not enabled; continuig...\n"));
					continue;
				}
				/* get Interrupt vector */
#ifndef KERNEL_2_4
				pci_read_config_byte( cspci_dev, PCI_INTERRUPT_LINE, 
									  &pci_irq_line );
				dbg_printf(("cs_scan: pci_irq_line = %d\n",pci_irq_line));
#else
				pci_irq_line = cspci_dev->irq;
				dbg_printf(("cs_scan: pci_irq_line = %d\n",pci_irq_line));
#endif
				/* get BaseAddress0 */
				pci_read_config_dword( cspci_dev, 
									   PCI_BASE_ADDRESS_0, 
									   &pci_ioaddr0 );
				/* Remove I/O space marker in bit 0. */
				pci_ioaddr0 &= ~3;

   				/* get BaseAddress1 even we might not use it */
				pci_read_config_dword( cspci_dev, 
									   PCI_BASE_ADDRESS_1, 
									   &pci_ioaddr1 );
				/* Remove I/O space marker in bit 0. */
				pci_ioaddr1 &= ~3;

   				/* Get and check the bus-master and latency values.
				   Some PCI BIOSes fail to set the master-enable bit, and
				   the latency timer must be set to the maximum value to avoid
				   data corruption that occurs when the timer expires during
				   a transfer.  Yes, it's a bug. */
				pci_read_config_word( cspci_dev, 
									  PCI_COMMAND, 
									  &pci_command );
				if (!(pci_command & PCI_COMMAND_MASTER))
				{
					dbg_printf(("PCI Master Bit has not been set! Setting...\n"));
					pci_command |= PCI_COMMAND_MASTER;
					pci_write_config_word( cspci_dev, 
										   PCI_COMMAND, 
										   pci_command );
				}
/*** sal
     pci_read_config_byte( cspci_dev, 
     PCI_LATENCY_TIMER, 
     &pci_latency );
     if ( pci_latency != 255 ) {
     dbg_printf(("Overriding PCI latency timer (CFLT) setting of %d,"
     " new value is 255.\n", pci_latency));
     pci_write_config_byte( cspci_dev,
     PCI_LATENCY_TIMER, 255);
     }
****/

#ifndef KERNEL_2_4
				cspci_dev->base_address[0] &= ~3;
				cspci_dev->base_address[1] &= ~3;
#endif

				/* probe the device if it's told */
				if (do_probe)
					rc = cs_found_device( cspci_dev );
				/* increment the number of card found if success */
				if (0 == rc)
					cards_found++;

			} /* while (! ... */
		} /* for (ver = 0; ... */
	} /* for (fun = 0; ... */

	return cards_found;
}



static int cs_found_device(struct pci_dev * pcid)
{
	struct cspci_device* dev = NULL;
	int rc;
	char* p;
	int size;
	int err = 0;

        DEVICE_PARAMS *pParams;

	/* find out what's the configuration for this device ID */
        GetDeviceData((pcid->vendor<<16) + pcid->device, &pParams);

#ifdef KERNEL_2_4
	dbg_printf(("Checking io resource 0.\n"));
	if (!(pci_resource_flags(pcid, 0) & IORESOURCE_IO))
		return -ENXIO;

	dbg_printf(("Checking free io resource 0.\n"));
	if ((err=check_region(pci_resource_start(pcid, 0), pci_resource_len(pcid, 0)))) {
		dbg_printf(("Checking free io resource 0 (rv=%d)!\n", err));
		return err;
	}

	if (pParams->flags & USE_OPREG) {
		dbg_printf(("Checking mem resource 1.\n"));
		if (!(pci_resource_flags(pcid, 1) & IORESOURCE_MEM))
			return -ENXIO;
		dbg_printf(("Checking free mem resource 1.\n"));
		if ((err=check_region(pci_resource_start(pcid, 1), pci_resource_len(pcid, 1)))) {
                        dbg_printf(("err = %d\n", err));
			return err;
                }
	}
	dbg_printf(("PCI resources available.\n"));
#endif

	/* Allocate and fill new device structure. */
	dev = (struct cspci_device *) kmalloc(sizeof(struct cspci_device), GFP_KERNEL);
	memset(dev, 0, sizeof(struct cspci_device));
	strcpy( dev->name, "cspci " );
	dev->name[strlen(dev->name)-1] = (char)num_cs_dev + '0';
	dev->opRegs = pci_resource_start(pcid, 1);
	dev->controlRegs = pci_resource_start(pcid, 0);
	dev->irq = pcid->irq;
	dev->stat.hardwareVersion = dev->device_id = pcid->device;
	dev->val_intclear = pParams->intclear;
	dev->flags |= BUS_MASTER_FLAG;       /* Bus master defaults to on */
	dev->flags |= pParams->flags;        /* see if we want to use OP reg */
	spin_lock_init(&dev->lock);

	/* safely concatenate two strings together */
	/* save one for the space in between */
	size = sizeof(dev->product_name) - 2; 
	p = dev->product_name;
	strncpy( p, pParams->versionName, size );
	strcat( p, " " );
	size -= strlen(p);
	p = strchr(p, 0);
	strncpy( p, pParams->versionName, size );

	/* add it to the array of device */
	rc = cs_probe1(dev);
	if (rc < 0)
	{
		kfree(dev);
		return rc;
	}

	dev->unit = num_cs_dev;
	root_cs_dev[num_cs_dev++] = dev;

	return err;
}

static int cs_probe1(struct cspci_device *dev)
{
	int error;

	printk("%s: %s\n", dev->name, dev->product_name);
	printk("IRQ %d,", dev->irq);
	if (dev->flags & USE_OPREG)
		printk(" Operational Regs at %4x, Control Regs at %4x\n",
			   dev->opRegs, dev->controlRegs);
	else
		printk(" Control Regs at %4x\n", dev->controlRegs);

  

	/*
	 * We must register the interrupt handler before calling
	 * cs_reset because cs_reset generates an interrupt.
	 */
	error = request_irq(dev->irq, cs_interrupt, SA_SHIRQ, 
			    dev->name, (void*)dev);

	if (error) 
	{
		printk("  Could not register IRQ (%d), error=%d\n", 
			dev->irq, error);
		free_irq(dev->irq, (void*)dev);
		return -ENODEV;
	}

	if (cs_reset(dev))
	{
		dbg_printf(("Device did not reset properly\n"));
		dev->flags |= OUT_OF_ORDER;
		free_irq(dev->irq, (void*)dev);
		return -ENXIO;
	}

  
	/* clear and enable the interrupt flag */
	CS_W_CONTROL(dev, REG_INTCONT, dev->val_intclear);

  	dbg_printf(("%s: request region\n", dev->name));
	/* We do a request_region() only to register /proc/ioports info. */
	if (dev->flags & USE_OPREG)
	{
		request_region(dev->opRegs, OPREG_SIZE, dev->name);
	}
	request_region(dev->controlRegs, CONTROLREG_SIZE, dev->name);
	dbg_printf(("request region dove\n"));

  	return 0;
}

static int cs_reset(struct cspci_device *dev)
{
	unsigned int HSValue;
	unsigned long start_time = 0;
	int found=0;
	/*unsigned int flags; cleared a warning due to this*/
	unsigned long flags;

	dbg_printf(("%s: start reset\n", dev->name));

	/*
	 * Let interrupt handler know that the interrupt came as a
	 * result of resetting the device.
	 */
	spin_lock_irqsave(&dev->lock, flags);
	dev->flags |= WAIT_FOR_RESET;
	spin_unlock_irqrestore(&dev->lock, flags);


	HSValue = CS_R_CONTROL(dev, REG_SIGNATURE);
	if (!(dev->flags & USE_OPREG) && (ISG_VENDOR_ID != (HSValue & 0xffff)))
	{
		printk("%s: invalid signature, %x\n", dev->name, (int)HSValue);
		return -ENXIO;
	}

/*	dbg_printf(("reg control b4 reset control=%x mailstat=%x\n",CS_R_CONTROL(dev, REG_CONTROL),CS_R_CONTROL(dev, REG_MAILSTAT)));
*/
	CS_W_CONTROL(dev, REG_CONTROL, VAL_RESET);
		 
	CS_W_CONTROL(dev, REG_CONTROL, VAL_CLEAR);

  	/* Wait for the card to get out of power-up selftest */
	if (dev->flags & USE_OPREG)
	{
		/* beta board */
		start_time = jiffies;
		do {
			HSValue = CS_R_OP(dev, OPREG_HANDSHAKE);
		} while (((jiffies - start_time) < TIME_OUT) &&
				 ((HSValue & VAL_CARDBUSY) != VAL_CARDBUSY));

		/* Start the card power-up selftest */
		CS_W_OP(dev, OPREG_HANDSHAKE, VAL_CLEAR);
	}

	/*
	 * Wait for card power-up selftest to finish.  I've found that
	 * it usually takes 500ms.  -mwang
	 */
	dbg_printf(("%s: start loop\n", dev->name));
	start_time = jiffies;
	mdelay(750);  /* wait 750 ms before checking */
	while (jiffies < start_time + TIME_OUT )
	{
		HSValue = CS_R_CONTROL(dev, REG_MAILSTAT); 
		if (((HSValue & VAL_ALL_MAIL_EMPTY) == VAL_ALL_MAIL_EMPTY) ||
			((dev->flags & WAIT_FOR_RESET) != WAIT_FOR_RESET))
		{
			found++;
			break;
		}
		udelay(1000);  /* wait 1 ms before checking again. */
	}

	dbg_printf(("%s: Mailbox registers are %x\n", dev->name, HSValue));
	dbg_printf(("elapsed jiffies %d\n", (int)(jiffies - start_time)));

	/* in all cases, clear the WAIT_FOR_RESET flag */
	spin_lock_irqsave(&dev->lock, flags);
	dev->flags &= ~WAIT_FOR_RESET;
	spin_unlock_irqrestore(&dev->lock, flags);

	if (found == 0)
	{
		printk("%s: reset timeout after %d secs, register=%x\n",
		       dev->name, ((int)(jiffies - start_time))/100, HSValue);
		return -ENXIO;
	}

	HSValue = CS_R_CONTROL(dev, REG_COMPCMD);

	dbg_printf(("CmdStat = %x\n", HSValue));
	if (dev->flags & USE_OPREG)
		dbg_printf(("State is %x\n", CS_R_OP(dev, OPREG_STATE)));

	/* clear the interrupt flag */
	CS_W_CONTROL(dev, REG_INTCONT, dev->val_intclear);

	CS_R_CONTROL(dev, REG_INTCONT);

  	/* reset the bus master control status reg */
	CS_W_CONTROL(dev, REG_CONTROL, VAL_BMRESET);

  	return HSValue;
}

static int cs_resizebuffer( HANDLEINFO* ph, int size )
{
	if (NULL != ph->orig_buf)
		kfree( ph->orig_buf );

	ph->bufsize = size;
	ph->orig_buf = (SW_BYTE*)kmalloc(ph->bufsize+4, GFP_KERNEL);
	if (NULL == ph->orig_buf) {
		ph->bufsize = 0;
		ph->buffer = NULL;
		return -ENOMEM;
	}

	/* make sure the buffer is on the next word boundry */
	if (((unsigned long)ph->orig_buf) & 0x3)
	{
		dbg_printf(("Shifting Buffer %x to 32bit boundry.\n", (int)ph->orig_buf));
	}
	ph->buffer  = (SW_BYTE *)((((unsigned long)ph->orig_buf)+3)&0xfffffffc);
	ph->dma_buf = (caddr_t)virt_to_bus(ph->buffer);  /* sal */
	return 0;
}

static int cs_open(struct inode * inode, struct file * file)
{
	HANDLEINFO *ph;

	dbg_printf(("Process %d is opening\n", (int)current->pid));

	/* we do not need the original inode */
	/* sal iput(inode);*/

#ifndef KERNEL_2_6
   	MOD_INC_USE_COUNT;
#endif
		file->private_data = 
		ph = (HANDLEINFO*)kmalloc(sizeof(HANDLEINFO), GFP_KERNEL); /*sal */
	if (ph == NULL) {
#ifndef KERNEL_2_6
		MOD_DEC_USE_COUNT;
#endif
		return -ENOMEM;
	}

	ph->orig_buf = NULL;
	if (cs_resizebuffer( ph, WINDOW_SIZE )) {
#ifndef KERNEL_2_6
		MOD_DEC_USE_COUNT;
#endif
		return -ENOMEM;
	}

	ph->flags = 0;
	ph->response = 0;
	init_waitqueue_head(&ph->waitq);


	dbg_printf(("Process %d got ph %x\n", (int)current->pid, (int)ph));
	return 0;
}


/*------------------------------------------------------------------------
 *
 * close routine.
 *
 *------------------------------------------------------------------------*/
static int cs_release(struct inode * inode, struct file * file)
{
	HANDLEINFO *ph = (HANDLEINFO*) file->private_data; /* sal */

	dbg_printf(("cs_release: handleinfo=0x%x\n", (int)ph));

	cs_cancel(ph);
	kfree(ph->orig_buf);
	kfree(ph);

#ifndef KERNEL_2_6
	MOD_DEC_USE_COUNT;
#endif
	return 0;
}


static inline int still_processing(HANDLEINFO *ph)
{
		return (ph->flags & HS_PROCESSING);
}

static int cs_ioctl(
	struct inode *inode, 
	struct file *file,
	unsigned int cmd, 
	unsigned long arg )
{
	HANDLEINFO *ph = (HANDLEINFO*)file->private_data;  /* sal */
	int retval = 0;
	unsigned long flags;
	int i;
	int memreq;
	CS_ADAPTERSTATS adpstat;

  	dbg_printf(("cs_ioctl: pid=%d, handleinfo=0x%x", current->pid, (int)ph));

	if (ph->flags & HS_CANCELPENDING)
		return -EBUSY;

	switch (cmd)
	{
	case CS_CMDGETRESPONSE:
		dbg_printf(("Top of CS_CMDGETRESPONSE\n"));

		/*
		 * Wait for the interrupt service routine to wake us up.
		 * Do not allow ourselves to catch an interrupt from the user while
		 * we are waiting.  Otherwise, the driver may free the DMA buffer
		 * while the DMA is still active.   -mwang 10/2/00
		 */
		dbg_printf(("handle is about to sleep, flags %x\n", ph->flags));
		wait_event(ph->waitq, !still_processing(ph));

        //init_waitqueue_head(&ph->waitq);
		if (ph->flags & HS_COMPLETE)
		{
			dbg_printf(("Setting response CS_CMDGETRESPONSE\n"));
			__copy_to_user(ph->cmd_info.resultBuffer, ph->buffer,
						   ph->cmd_info.resultSize);
			__copy_to_user((unsigned long*)arg, &ph->response, 
						   sizeof(unsigned long));
			dbg_printf(("Returning from CS_CMDGETRESPONSE\n"));
			ph->flags &= ~HS_COMPLETE;
			retval = 0;
		} else
		{
			dbg_printf(("No responses waiting\n"));
			retval = -ENODATA;
		}
		break;

	case CS_CMDEXEC:
		dbg_printf(("Top of CS_CMDEXEC\n"));

		if (ph->flags & (HS_PROCESSING | HS_COMPLETE))
		{
			dbg_printf(("Pending command already active\n"));
			return -EBUSY;
		}
		ph->flags |= HS_PROCESSING;

		__copy_from_user(&ph->cmd_info, (CS_CMDINFO*)arg, sizeof(CS_CMDINFO));
		memreq = ph->cmd_info.dataSize > ph->cmd_info.resultSize ?
			ph->cmd_info.dataSize :
				ph->cmd_info.resultSize;
		if (ph->bufsize < memreq)
		{
			dbg_printf(("resizing buffer from %d to %d\n", ph->bufsize, memreq));
			if (cs_resizebuffer( ph, memreq ) < 0)
			{
				ph->flags &= ~HS_PROCESSING;
				return -ENOMEM;
			}
		}
		__copy_from_user(ph->buffer, ph->cmd_info.dataBuffer,ph->cmd_info.dataSize);
		dbg_printf(("Queueing request\n"));
		dbg_printf(("dataSize %d, resultSize %d\n", (int)ph->cmd_info.dataSize, 
					(int)ph->cmd_info.resultSize));
		dbg_printf(("Going to adapter %x\n", (int)ph->cmd_info.accnum));
		if ((ANY_ACCELERATOR != ph->cmd_info.accnum) &&
			(ph->cmd_info.accnum >= num_cs_dev))
		{
			dbg_printf(("Adapter %d is not a valid adapter, total %d\n",
						(int)ph->cmd_info.accnum, num_cs_dev));
			retval = -ENXIO;
		} else
		{
			cs_queuereq(ph);
			cs_processreq(ph->cmd_info.accnum);
			dbg_printf(("CS_CMDEXEC: Done processing request\n"));
		}
		break;

	case CS_CMDGETGLOBALSTATS:
		/* calculate # of interrupt received */
		spin_lock_irqsave(&cs_glock, flags);
		stat.interruptReceived = 0;
		for (i = 0; i < num_cs_dev; ++i)
			stat.interruptReceived += root_cs_dev[i]->stat.interruptReceived;
		spin_unlock_irqrestore(&cs_glock, flags);
		__copy_to_user((CS_GLOBALSTATS*)arg, &stat, sizeof(stat));
		break;

	case CS_CMDGETADAPTERSTATS:
		__copy_from_user((void*)&adpstat, (CS_ADAPTERSTATS*)arg, 
						 sizeof(CS_ADAPTERSTATS));
		if (adpstat.accnum >= num_cs_dev)
			retval = -EINVAL;
		else
			__copy_to_user((void*)arg, &(root_cs_dev[adpstat.accnum]->stat), 
						   sizeof(CS_ADAPTERSTATS));
		break;

#ifdef CS_DEBUG
	case 0x106:
		cs_debug();
		break;
#endif

	default:
		retval = -EINVAL;
	}
	return retval;
}


static void cs_processreq(unsigned long accnum)
{
	struct cspci_device *dev;
	HANDLEINFO *ph;

  	dbg_printf(("Top of process request routine\n"));

	/* Try to get a device first. */
	dev = cs_requestdev(accnum);
	if (dev == NULL) {
		dbg_printf(("No available devices in the cs_processreq routine!!\n"));
		return;
	}

  	ph = cs_removehead(dev);
	if (ph == NULL) {
		cs_freedev(dev);
		return;
	}

  	dbg_printf(("About to process request\n"));
	cs_executereq(dev, ph);
}

/* This is our bottom half routine */
#if CS_BROKEN_BH
static void cs_bottomentry(void)
#else
static void cs_bottomentry(void *userdata)
#endif
{
	int i;
	unsigned long flags;

	/* we are running now, so someone will have to schedule us again */
	bh_scheduled = 0;

	for (i = 0; i < num_cs_dev; ++i)
	{
		if (root_cs_dev[i]->flags & NEED_DPC)
		{
			spin_lock_irqsave(&(root_cs_dev[i]->lock), flags);
			root_cs_dev[i]->flags &= ~NEED_DPC;
			spin_unlock_irqrestore(&(root_cs_dev[i]->lock), flags);
			cs_dpc( root_cs_dev[i] );
		}
	}
}

/* This is our bottom half routine */
static void cs_dpc(struct cspci_device* dev)
{
	HANDLEINFO *ph;
	unsigned int cmd_index;
	unsigned long flags;

  	dbg_printf(("Top of dpc routine\n"));

	/* A process may be sleeping waiting for the result of the last command.
	   We will wake the process up here. */
	ph = dev->ph;
	if (ph != NULL)
	{
		ph->response = dev->response;
		ph->flags |= HS_COMPLETE;
		ph->flags &= ~HS_PROCESSING;
		dbg_printf(("Cleared HS_PROCESSING flag, %x\n", ph->flags));

		/* update the statistics */
		cmd_index = ph->cmd_info.command;
		if (cmd_index > (MAX_CSCMD-1))
			cmd_index = MAX_CSCMD - 1;

		spin_lock_irqsave(&dev->lock, flags);
		++dev->stat.requestCompleted;
		++dev->stat.cmdstat[cmd_index].requestCompleted;
		spin_unlock_irqrestore(&dev->lock, flags);

		spin_lock_irqsave(&cs_glock, flags);
		++stat.requestCompleted;
		++stat.cmdstat[cmd_index].requestCompleted;
		spin_unlock_irqrestore(&cs_glock, flags);

		if (waitqueue_active(&ph->waitq)) {
			/* someone is indeed waiting for the command*/
			dbg_printf(("Waking up the sleeping command\n"));
			wake_up(&ph->waitq);
		}
	}

	/* dequeue the next one if there is an one */
	ph = cs_removehead(dev);
	if (NULL != ph)
	{
		dbg_printf(("Chaining: to process request\n"));
		cs_executereq(dev, ph);
	} else
	{
		cs_freedev(dev);
	}
}


/*
 * At this point, we have a free and locked device and we have a
 * request.  Program the hardware to DMA and process the request.
 * This function must be called with no locks held.
 */
static void cs_executereq(struct cspci_device *dev, HANDLEINFO *ph)
{
	unsigned long flags;
	unsigned int cmd_index;
	SW_U32 sizes = ((ph->cmd_info.resultSize)<< 16) | 
		((ph->cmd_info.dataSize) & 0xFFFF);

	dbg_printf(("dataSize is %d\n", (int)ph->cmd_info.dataSize));
	dbg_printf(("resultSize is %d\n", (int)ph->cmd_info.resultSize));
	dbg_printf(("sizes are %x\n", (int)sizes));
	dbg_printf(("command is 0x%04x\n", (int)ph->cmd_info.command));

	cmd_index = ph->cmd_info.command;
	if (cmd_index > (MAX_CSCMD-1))
	    cmd_index = MAX_CSCMD - 1;

	/* Associate the device with the process */
	spin_lock_irqsave(&dev->lock, flags);
	dev->ph = ph;
	++dev->stat.requestAttempted;
	++dev->stat.cmdstat[cmd_index].requestAttempted;
	spin_unlock_irqrestore(&dev->lock, flags);


	spin_lock_irqsave(&cs_glock, flags);
	++stat.requestAttempted;
	++stat.cmdstat[cmd_index].requestAttempted;

	/* Write to the Address and count registers for the fifo */
	if (dev->flags & USE_FASTMAP_IOREG)
	{
		dbg_printf(("Use FASTMAP_IOREG option\n"));
		CS_W_CONTROL(dev, REG_FASTMAP_TO, (unsigned long)ph->dma_buf); /* sal */
		CS_W_CONTROL(dev, REG_FASTMAP_FROM, (unsigned long)ph->dma_buf); /* sal */
	}

	CS_W_CONTROL(dev, REG_TO_ADDR, (unsigned long)ph->dma_buf); /* sal */
	CS_W_CONTROL(dev, REG_TO_SIZE, ph->cmd_info.resultSize);
	CS_W_CONTROL(dev, REG_FROM_ADDR, (unsigned long)ph->dma_buf); /*sal*/
	CS_W_CONTROL(dev, REG_FROM_SIZE, ph->cmd_info.dataSize);
	CS_W_CONTROL(dev, REG_SIZES, sizes);

 	printhex("Output buffer before: ", ph->buffer, ph->cmd_info.resultSize); 
	/* Enable the fifo */
	CS_W_CONTROL(dev, REG_CONTROL, VAL_BMENABLE);

    /* enable MBX reg 4 Write interrupt to host from CS3 */
    if (dev->device_id == CS3_DEVICE_ID)
        CS_W_CONTROL(dev, REG_INTCONT, VAL_CS3_ICR_INTENABLE);

	/* Write the command number to start the command */
	CS_W_CONTROL(dev, REG_COMMAND, ph->cmd_info.command);

	printhex("Output buffer after: ", ph->buffer, ph->cmd_info.resultSize); 
	spin_unlock_irqrestore(&cs_glock, flags);
}


/*
 * Acquire the device specified by the accnum.
 * This function must be called with no locks held.
 * Returns NULL if the device was not acquired, or the locked device
 * if it was acquired.
 */
static struct cspci_device* cs_requestdev(unsigned long accnum)
{
	struct cspci_device *dev, *ret_dev;
	int i;
	int found;
	unsigned long flags;

	if (ANY_ACCELERATOR != accnum)
	{
		dbg_printf(("Request specific device %d\n", (int)accnum));
		dev = root_cs_dev[accnum];
		dbg_printf(("dev %x, dev->flags %x\n", (int)dev, (int)dev->flags));

		if (dev == NULL)
			return NULL;

		spin_lock_irqsave(&dev->lock, flags);
		if (!(dev->flags & CARDBUSY) && !(dev->flags & OUT_OF_ORDER))
		{
			dev->flags |= CARDBUSY;
			ret_dev = dev;
		}
		else
		{
			ret_dev = NULL;
		}
		spin_unlock_irqrestore(&dev->lock, flags);

		return ret_dev;
	}

	dbg_printf(("Request any device\n"));
	i = next_cs_dev;
	found = 0;
	do {
		dev = root_cs_dev[i];
		if (dev != NULL)
		{
			spin_lock_irqsave(&dev->lock, flags);
			if (!(dev->flags & CARDBUSY) && !(dev->flags & OUT_OF_ORDER))
			{
				dbg_printf(("found idle device %d, flag %x\n", i, dev->flags));
				found = 1;
				dev->flags |= CARDBUSY;
			}
			spin_unlock_irqrestore(&dev->lock, flags);
			if (found)
				break;
		}
		i = (i+1) % num_cs_dev;
	} while (i != next_cs_dev);

	if (0 == found)
		return NULL;

	next_cs_dev = (i+1) % num_cs_dev;
	return dev;
}


/*
 * Mark the device as free.
 * This function must be called with no locks held.
 */
static void cs_freedev(struct cspci_device* dev)
{
	unsigned long flags;

	spin_lock_irqsave(&dev->lock, flags);
	dev->ph = NULL;
	dev->flags &= ~(CARDBUSY);
	spin_unlock_irqrestore(&dev->lock, flags);

	dbg_printf(("cs_freedev: 0x%x flags=0x%x\n",
		(int)dev, (int)dev->flags));
}


/*
 * Remove a request from either the per adapter queue or from
 * the global queue.
 * This function must be called with no locks held.
 * Returns a request if there is one, otherwise, NULL.
 */
static HANDLEINFO* cs_removehead(struct cspci_device* dev)
{
	HANDLEINFO* ph;
	unsigned long flags;

	/* check per device queue */
	ph = dev->wait_head;
	if (NULL != ph)
	{
		/* This is a bit tricky, we have to check twice 
		 * because on the first check, we did not have a lock.
		 * So two threads could have both seen the request,
		 * but only one can have it.  That's why we get the lock
		 * and check again.  Most of the time, the top check
		 * will fail since libswift queues requests on the global
		 * list.  -mwang 10/6/2000
		 */
		spin_lock_irqsave(&dev->lock, flags);
		ph = dev->wait_head;
		if (NULL != ph)
		{ 
			dev->wait_head = ph->next;
			if (ph == dev->wait_tail)
				dev->wait_tail = NULL;
		}
		spin_unlock_irqrestore(&dev->lock, flags);

		if (NULL != ph)
		{
			spin_lock_irqsave(&cs_glock, flags);
			--stat.currentPendingSize;
			spin_unlock_irqrestore(&cs_glock, flags);

			ph->flags &= ~HS_CANCELLABLE;
			return ph;
		}
	}

	/*
	 * If we got to this point, then we have not found a
	 * request yet.  Look in the global queue.
	 */
	spin_lock_irqsave(&cs_glock, flags);

	ph = wait_head;
	/* get the next item off the queue */
	if (NULL != ph)
	{
	    wait_head = ph->next;
	    if (ph == wait_tail)
	      wait_tail = NULL;
	    --stat.currentPendingSize;
	}

	spin_unlock_irqrestore(&cs_glock, flags);

	/* request CANNOT be cancelled */
	if (ph)
		ph->flags &= ~HS_CANCELLABLE;

	return ph;
}


/*
 * Queue the request on either a device queue or in the global queue.
 * This function must be called with no locks held.
 */
static void cs_queuereq(HANDLEINFO *ph) 
{
	unsigned long flags;

	/*
	 * First grab the global spin lock and update the global
	 * statistics.  Then see if we are queueing this request
	 * on the global queue.  If so, we can do that since we
	 * already have the global lock.  (Saves us from having to
	 * grab and release it twice.)		-mwang 10/6/2000
	 */
	spin_lock_irqsave(&cs_glock, flags);

	stat.sumOfPendingSize += stat.currentPendingSize;
	++stat.currentPendingSize;
	if (stat.currentPendingSize > stat.maxPendingSize)
		stat.maxPendingSize = stat.currentPendingSize;


	if (ANY_ACCELERATOR == ph->cmd_info.accnum)
	{
	    /* put this request into the global request queue.
	     * Note that we already have the global lock at this point */

		ph->next = NULL;
		if (NULL == wait_head)
			wait_head = ph;

		if (NULL != wait_tail)
			wait_tail->next = ph;

		wait_tail = ph;

	    spin_unlock_irqrestore(&cs_glock, flags);
	}
	else
	{
	    /* Put the request on the queue of a specific device.
	     * We assume that the accnum is valid - check in cs_ioctl.
	     * We have to release the global lock first, and then
	     * acquire the device lock.
	     */
	  struct cspci_device* dev;

	  spin_unlock_irqrestore(&cs_glock, flags);

	  dev = root_cs_dev[ph->cmd_info.accnum];

	  spin_lock_irqsave(&dev->lock, flags);

	  ph->next = NULL;
	  if (NULL == dev->wait_head)
	      dev->wait_head = ph;

      if (NULL != dev->wait_tail)
	      dev->wait_tail->next = ph;

	  dev->wait_tail = ph;

	  spin_unlock_irqrestore(&dev->lock, flags);
	}

	/* request can be cancelled */
	ph->flags |= HS_CANCELLABLE;
}


/*
 * Dequeue the specified request from either the global queue or
 * from a device queue.  This function is called when the file
 * descriptor is closed.  (Hmm, I'm not sure this will ever happen,
 * but I'll fix it anyways.)
 * This function must be called with no locks held.
 * Returns 0 if the request was found, -1 otherwise.
 */
static int cs_dequeuereq(HANDLEINFO *ph) 
{
	unsigned long flags;
	HANDLEINFO* p;
	struct cspci_device* dev;
	int removed=0;
	int i;

	/* try to find the request on the global queue first */
	spin_lock_irqsave(&cs_glock, flags);

	if (ph != wait_head)
	{
		for (p = wait_head; p; p = p->next)
		{
			if (p->next == ph)
			{
				removed = 1;
				p->next = ph->next;
				if (wait_tail == ph)
					wait_tail = p;
				break;
			}
		}
	} 
	else
	{
		removed = 1;
		wait_head = ph->next;
		if (wait_tail == ph)
			wait_tail = NULL;
	}

	/*
	 * If we found the request on the global list, then we are done.
	 * Just return from here.
	 */
	if (removed == 1)
	{
	    --stat.currentPendingSize;
	    spin_unlock_irqrestore(&cs_glock, flags);

	    ph->flags &= ~HS_CANCELLABLE;
	    return 0;
	}

	spin_unlock_irqrestore(&cs_glock, flags);


	/*
	 * Now go through each of the devices and look for the
	 * request on its queue.
	 */
	for (i = 0; i < num_cs_dev; ++i)
	{
		dev = root_cs_dev[i];
		if (dev == NULL)
			continue;

		spin_lock_irqsave(&dev->lock, flags);
		if (ph != dev->wait_head)
		{
			for (p = dev->wait_head; p; p = p->next)
			{
				if (p->next == ph)
				{
					removed = 1;
					p->next = ph->next;
					if (wait_tail == ph)
						wait_tail = p;
					break;
				}
			}
		}
		else
		{
			removed = 1;
			dev->wait_head = ph->next;
			if (wait_tail == ph)
				wait_tail = NULL;
		}
		spin_unlock_irqrestore(&dev->lock, flags);

		if (removed == 1)
			break;
	}


	if (removed == 1)
	{
		/* request CANNOT be cancelled */
		ph->flags &= ~HS_CANCELLABLE;

		spin_lock_irqsave(&cs_glock, flags);
		--stat.currentPendingSize;
		spin_unlock_irqrestore(&cs_glock, flags);
		return 0;
	}

	return -1;
}
#ifdef KERNEL_2_6
static irqreturn_t cs_interrupt(int irq, void *dev_id, struct pt_regs *regs)
#else
static void cs_interrupt(int irq, void *dev_id, struct pt_regs *regs)
#endif
{
	int status;
	struct cspci_device *dev;
	unsigned long flags;
	int do_queue=0;

	dev = (struct cspci_device*) dev_id;


	/* check if the interrupt is sent to a valid board */
	/* We need to find the board that caused the interrupt */
	status = CS_R_CONTROL(dev, REG_INTCONT);

	if (0 == (status & VAL_INTERRUPTED))
	{
		/*dbg_printf(("The interrupt was not for a CSPCI device\n"));*/
#ifdef KERNEL_2_6
		return IRQ_NONE; 
#else
		return;
#endif
	} else
	{
		dbg_printf(("Got interrupt for board, int val %x\n", status));
	}

	spin_lock_irqsave(&cs_glock, flags);
	++stat.interruptServiced;
	spin_unlock_irqrestore(&cs_glock, flags);

	spin_lock_irqsave(&dev->lock, flags);
	++dev->stat.interruptReceived;

	/*
	 * If this interrupt was the result of resetting the board,
	 * clear the flag and return.  
	 */
	if (dev->flags & WAIT_FOR_RESET)
	{
	    dbg_printf(("cs_interrupt: clearin WAIT_FOR_RESET\n"));
	    dev->flags &= ~WAIT_FOR_RESET;
	    spin_unlock_irqrestore(&dev->lock, flags);

	    /* clear the interrupt in hardware */
	    CS_W_CONTROL(dev, REG_INTCONT, dev->val_intclear);
	    status = CS_R_CONTROL(dev, REG_INTCONT);
#ifdef KERNEL_2_6
		return IRQ_HANDLED; 
#else
		return;
#endif
	}

	if (0 == (dev->flags & CARDBUSY))
	{
	  	spin_unlock_irqrestore(&dev->lock, flags);

		dbg_printf(("This board is not currently in use\n"));
		/* clear the interrupt flag */
		CS_W_CONTROL(dev, REG_INTCONT, dev->val_intclear);

    	status = CS_R_CONTROL(dev, REG_INTCONT);
		dbg_printf(("Cleard the interrupt, int val %x\n", status));

#ifdef KERNEL_2_6
		return IRQ_HANDLED; 
#else
		return;
#endif
	}

	++dev->stat.interruptServiced;

	if (dev->flags & INTERRUPT)
	{
		dbg_printf(("%s: Re-entering the cs_interrupt handler.\n",dev->name));
	}
	dev->flags |= INTERRUPT;

  
	/* Read the command status register for the host */
	dev->response = CS_R_CONTROL(dev, REG_COMPCMD);
	dbg_printf(("Response for command %x\n", dev->response));

	/* clear the interrupt flag */
	CS_W_CONTROL(dev, REG_INTCONT, dev->val_intclear);
	status = CS_R_CONTROL(dev, REG_INTCONT);

	/* reset the bus master control status reg */
	CS_W_CONTROL(dev, REG_CONTROL, VAL_BMRESET);

  	/* Free up this device for a waiting command */
	dev->flags &= ~(INTERRUPT);

	/* A process may be sleeping waiting for the result of the last command.
	   We will wake the process up here. */
	if (bh_csno >= 0)
	{
		dev->flags |= NEED_DPC;
		spin_unlock_irqrestore(&dev->lock, flags);

	#if ! CS_BROKEN_BH
		/* If our delayed processing function has not been
		* scheduled yet, schedule it.
		*/
		spin_lock_irqsave(&cs_glock, flags);
		if (bh_scheduled == 0)
		{
			bh_scheduled = 1;
			do_queue = 1;
		}
		spin_unlock_irqrestore(&cs_glock, flags);
		if (do_queue == 1)
#ifdef KERNEL_2_6
			schedule_work(&cs_task);
#else 
			queue_task(&cs_task, &tq_immediate);
#endif
#endif

#ifndef KERNEL_2_6
		mark_bh(bh_csno);
#endif
	} else
	{
		/* no bottom half */
		spin_unlock_irqrestore(&dev->lock, flags);
		cs_dpc(dev);
	}
#ifdef KERNEL_2_6
		return IRQ_HANDLED; 
#else
		return;
#endif
}


static void cs_cancel(HANDLEINFO* ph)
{
	/* a request is in processing */
	if (ph->flags & HS_PROCESSING)
	{
		/* request is in the queue and can be removed */
		if (ph->flags & HS_CANCELLABLE)
		{
			int rc;
			/* search all the queues to remove it */

			//printk("cs_cancel: ph(%p), wait_head(%p), wait_tail(%p)\n", ph, wait_head, wait_tail);
			rc = cs_dequeuereq(ph);
			//printk("cs_cancel(%d): ph(%p), wait_head(%p), wait_tail(%p)\n", rc, ph, wait_head, wait_tail);
		}

		/* request is not cancellable */
		else
		{
			/* This request can be either in following states:
			 * (1) just dequeued
			 * (2) being issued to the adapter
			 * (3) to be serviced by ISR
			 * (4) to be processed by DPC
			 * The best bet is to wait for it finish
			 */

     		//printk("cs_cancel: ph(0x%p) is being processed\n", ph);
			ph->flags |= HS_CANCELPENDING;
			//printk("cs_cancel: before sleep: ph(0x%p)->flags = 0x%x\n", ph, ph->flags);
			sleep_on(&ph->waitq);
			init_waitqueue_head(&ph->waitq);
			//printk("cs_cancel: after sleep: ph(0x%p)->flags = 0x%x\n", ph, ph->flags);

		}

	} /* if (ph->flags & HS_PROCESSING) */

	/* wake up any pending process */
	if (waitqueue_active(&ph->waitq))
	{
		wake_up(&ph->waitq);
	}
}

#ifdef CS_DEBUG
static void
cs_debug( void )
{
	struct cspci_device *dev;
	int i;

	printk("wait_head=%p, wait_tail=%p, pending=%ld\n", 
		   wait_head, wait_tail, stat.currentPendingSize );
	for (i = 0; i < num_cs_dev; ++i)
	{
		dev = root_cs_dev[i];
		printk("dev(%p): flags=0x%08x, ph=%p, wait_head=%p, wait_tail=%p\n", 
			   dev, dev->flags, dev->ph, dev->wait_head, dev->wait_tail);
	}
}
#endif

static void printhex(unsigned char *message, unsigned char *p, unsigned int size)

{
#ifdef CS_DEBUG
	unsigned int i;
	unsigned char *pc;

	printk("\n%s", message);

	for (i = 0, pc = p; i < size; i++, pc++)
	{
		printk("%02x ", *pc);
	}

	printk("\n");
#endif
}


