/*
 *  Copyright (c) 1997 Rainbow Technologies Inc, Internet Security Group. 
 *  All rights reserved.
 *
 *  Copyright (C) 2000 Paolo Molaro <lupus@linuxcare.com>. All rights reserved.
 *  Copyright (C) 2000 Marco Pantaleoni <panta@linuxcare.com>. All rights reserved.
 *
 *  @(#)cspci.h     1.0.1    3/07/97
 *
 *  Description:    headers for CryptoSwift kernel mode driver
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL RAINBOW
 * TECHNOLOGIES BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN 
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION 
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *  
 * This source code is subject to the U.S. Export Administration Regulations and 
 * other U.S. law, and may not be exported or re-exported to certain countries 
 * (currently Afghanistan (Taliban-controlled areas), Cuba, Iran, Iraq, Libya, 
 * North Korea, Serbia (except Kosovo), Sudan and Syria) or to persons or entities 
 * prohibited from receiving U.S. exports (including Denied Parties, entities on 
 * the Bureau of Export Administration Entity List, and Specially Designated 
 * Nationals). 
 *  
 * Permission is hereby granted, free of charge, to any person meeting the export 
 * regulations above to deal in this software and associated documentation files 
 * (the "Software"), without restriction, including without limitation the rights 
 * to use, copy, modify, and merge copies of the Software, and to permit persons 
 * to whom the Software is furnished to do so, subject to the following conditions:
 *  
 * 1.  The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 * 
 */

/* $Id: cspci.h,v 1.1.1.1 2002/10/01 02:55:50 jgagne Exp $ */

#ifndef __CSPCI_H__
#define __CSPCI_H__

#include "swiftype.h"
#include "swifterr.h"

/* IOCTL Commands */
#if defined(__FreeBSD__)
#include <sys/ioctl.h>
#define CS_CMDEXEC				_IOW('w',0x40,CS_CMDINFO)
#define CS_CMDGETRESPONSE		_IOW('w',0x41,SW_U32)
#define CS_CMDGETSTAT			_IOR('w',0x42,CS_CMDSTAT)
#define CS_CMDGETCONTEXT		_IOR('w',0x43,int)
#define CS_CMDGETGLOBALSTATS    _IOR('w',0x44,CS_GLOBALSTATS)
#define CS_CMDGETADAPTERSTATS	_IOR('w',0x45,CS_ADAPTERSTATS)
#define CS_CMDSETAPPSTORAGE		_IOW('w',0x46,)
#elif defined(LINUX)
#define CS_CMDEXEC				0x101
#define CS_CMDGETRESPONSE		0x102
#define CS_CMDGETGLOBALSTATS	0x103
#define CS_CMDGETADAPTERSTATS	0x104
#define CS_CMDSETAPPSTORAGE		0x105
#endif

#define CS_MAJOR				61

#if defined(WIN32)
#if !defined(_NTDDK_)
#include <winioctl.h>
#endif /* _NTDDK_ */
#define CSPCI_WIN32_NAME		"\\\\.\\CSPCI"

#define CS_CMDEXEC \
   CTL_CODE (FILE_DEVICE_UNKNOWN, 0xAA5, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define CS_GETGLOBALSTATS \
   CTL_CODE (FILE_DEVICE_UNKNOWN, 0xAA6, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define CS_GETADAPTERSTATS \
   CTL_CODE (FILE_DEVICE_UNKNOWN, 0xAA7, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define CS_SETAPPSTORAGE \
   CTL_CODE (FILE_DEVICE_UNKNOWN, 0xAA8, METHOD_BUFFERED, FILE_ANY_ACCESS)
#endif /* WIN32 */

#define ANY_ACCELERATOR			(0xffffffffL)

#define MAX_CSCMD				32

#define MAX_APP_STORAGE			16

typedef struct
{
	SW_U32   command;
    SW_U32   dataSize;
    SW_BYTE *dataBuffer;          
    SW_U32   resultSize;
    SW_BYTE *resultBuffer;        
    SW_U32   status;
    SW_U32   accnum;
} CS_CMDINFO;

typedef struct
{
	SW_U32 requestAttempted;
    SW_U32 requestCompleted;
} CS_CMDSTATS;

typedef struct
{
	SW_U32 driverVersion;          
    SW_U32 numAccelerator;
    SW_U32 interruptServiced;   
    SW_U32 interruptReceived;   
    SW_U32 requestAttempted;    
    SW_U32 requestCompleted;    
    SW_U32 maxPendingSize;         
    SW_U32 currentPendingSize;
    SW_U32 sumOfPendingSize;
    SW_U32 totalPendingCancel;       
    CS_CMDSTATS cmdstat[MAX_CSCMD]; 
    SW_U32 setByApp[16];				/* set and used by application */
} CS_GLOBALSTATS;

typedef struct {
    SW_U32 hardwareVersion;				/* version of hardware, RID from PCI    */
    SW_U32 interruptServiced;
    SW_U32 interruptReceived;
    SW_U32 requestAttempted;
    SW_U32 requestCompleted;
    CS_CMDSTATS cmdstat[MAX_CSCMD]; 
    SW_U32 setByApp[MAX_APP_STORAGE];	/* set and used by application */
    SW_U32 accnum;
} CS_ADAPTERSTATS;

typedef struct {
    SW_U32  accnum;
    SW_U32  startidx;
    SW_U32  numitem;
    SW_U32  data[1];
} CS_APPSTORAGE;
#define CS_APPSTORAGE_SIZE(p) (sizeof(CS_APPSTORAGE)+((p)->numitem-1)*sizeof(SW_U32))

#define CS_VERSION_CODING(maj,min,bld)  (((maj)&0x000000ff)<<24 |  ((min)&0x000000ff)<<16 | ((bld)&0x0000ffff))

#define CS_VERSION_MAJOR(v)             (((v)&0xff000000L)>>24)
#define CS_VERSION_MINOR(v)             (((v)&0x00ff0000L)>>16)
#define CS_VERSION_BUILD(v)             ((v)&0x0000ffffL)


#endif /* __CSPCI_H__ */
